# coding=utf-8
from __future__ import unicode_literals

import ujson
from rest_framework.compat import six
from rest_framework.exceptions import ParseError
from rest_framework.renderers import JSONRenderer


class BaseParser(object):
    def parse(self, data):
        """
        Given a stream to read from, return the parsed representation.
        Should return parsed data, or a `DataAndFiles` object consisting of the
        parsed data and files.
        """
        raise NotImplementedError(".parse() must be overridden.")


class UJSONParser(BaseParser):
    renderer_class = JSONRenderer

    def parse(self, data):
        try:
            return ujson.loads(data)
        except ValueError as exc:
            raise ParseError('JSON parse error - %s' % six.text_type(exc))



# coding=utf-8
from __future__ import unicode_literals

from django.conf import settings
from django.test.signals import setting_changed
from django.utils import six
from rest_framework.compat import importlib

DEFAULTS = {
    # Base API policies
    'DEFAULT_TEXT_RENDERER_CLASS': 'rest_channels.renderers.UJSONRenderer',
    'DEFAULT_BYTES_RENDERER_CLASS': None,
    'DEFAULT_TEXT_PARSER_CLASS': 'rest_channels.parsers.UJSONParser',
    'DEFAULT_BYTES_PARSER_CLASS': None,
    # Exception handling
    'EXCEPTION_HANDLER': 'rest_channels.views.exception_handler',
    'NON_FIELD_ERRORS_KEY': 'non_field_errors',
}

# List of settings that may be in string import notation.
IMPORT_STRINGS = (
    'DEFAULT_TEXT_RENDERER_CLASS',
    'DEFAULT_BYTES_RENDERER_CLASS',
    'DEFAULT_TEXT_PARSER_CLASS',
    'DEFAULT_BYTES_PARSER_CLASS',
    'EXCEPTION_HANDLER',
    'NON_FIELD_ERRORS_KEY'
)


def perform_import(val, setting_name):
    """
    If the given setting is a string import notation,
    then perform the necessary import or imports.
    """
    if val is None:
        return None
    elif isinstance(val, six.string_types):
        return import_from_string(val, setting_name)
    elif isinstance(val, (list, tuple)):
        return [import_from_string(item, setting_name) for item in val]
    return val


def import_from_string(val, setting_name):
    """
    Attempt to import a class from a string representation.
    """
    try:
        # Nod to tastypie's use of importlib.
        parts = val.split('.')
        module_path, class_name = '.'.join(parts[:-1]), parts[-1]
        module = importlib.import_module(module_path)
        return getattr(module, class_name)
    except (ImportError, AttributeError) as e:
        msg = ("Could not import '%s' for API setting '%s'. %s: %s." %
               (val, setting_name, e.__class__.__name__, e))
        raise ImportError(msg)


class RESTChannelsSettings(object):
    def __init__(self, user_settings=None, defaults=None, import_strings=None):
        if user_settings:
            self._user_settings = user_settings
        self.defaults = defaults or DEFAULTS
        self.import_strings = import_strings or IMPORT_STRINGS

    @property
    def user_settings(self):
        if not hasattr(self, '_user_settings'):
            self._user_settings = getattr(settings, 'REST_CHANNELS', {})
        return self._user_settings

    def __getattr__(self, attr):
        if attr not in self.defaults:
            raise AttributeError("Invalid REST_CHANNELS setting: '%s'" % attr)

        try:
            # Check if present in user settings
            val = self.user_settings[attr]
        except KeyError:
            # Fall back to defaults
            val = self.defaults[attr]

        # Coerce import strings into classes
        if attr in self.import_strings:
            val = perform_import(val, attr)

        # Cache the result
        setattr(self, attr, val)
        return val


rest_channels_settings = RESTChannelsSettings(None, DEFAULTS, IMPORT_STRINGS)


def reload_settings(*args, **kwargs):
    global rest_channels_settings
    setting, value = kwargs['setting'], kwargs['value']
    if setting == 'REST_CHANNELS':
        rest_channels_settings = RESTChannelsSettings(value, DEFAULTS, IMPORT_STRINGS)


setting_changed.connect(reload_settings)



from __future__ import unicode_literals

import six
import sys


class Empty(object):
    """
    Placeholder for unset attributes.
    Cannot use `None`, as that may be a valid value.
    """
    pass


def _hasattr(obj, name):
    return not getattr(obj, name) is Empty


class ContentType(object):
    text = 'text'
    bytes = 'bytes'
    none = 'none'


class SocketRequest(object):
    def __init__(self, message, text_parser_class=None, bytes_parser_class=None):
        self._message = message
        self.text_parser_class = text_parser_class
        self.bytes_parser_class = bytes_parser_class
        self._text = Empty
        self._bytes = Empty
        self._full_data = Empty

    @property
    def content_type(self):
        if self._message.content.get(ContentType.text) is not None:
            return ContentType.text
        elif self._message.content.get(ContentType.bytes) is not None:
            return ContentType.bytes
        else:
            return ContentType.none

    @property
    def data(self):
        if not _hasattr(self, '_full_data'):
            self._load_text_and_bytes()
        return self._full_data

    def _load_text_and_bytes(self):
        """
        Parses the message content into `self.data`.
        """
        if not _hasattr(self, '_bytes'):
            self._text, self._bytes = self._parse()
            if self._bytes:
                self._full_data = self._text.copy()
                self._full_data.update(self._bytes)
            else:
                self._full_data = self._text

    def select_parser(self):
        if self.content_type == ContentType.text:
            return self.text_parser_class
        else:
            return self.bytes_parser_class

    def _parse(self):
        media_type = self.content_type
        parser = self.select_parser()
        if parser is not None:
            try:
                parsed = parser().parse(self._message.content.get(media_type))
                if media_type == ContentType.text:
                    return parsed, None
                elif media_type == ContentType.bytes:
                    return None, parsed
                else:
                    return None, None
            except:
                self._text = None
                self._bytes = None
                raise
        return None, None

    def __getattribute__(self, attr):
        """
        If an attribute does not exist on this instance, then we also attempt
        to proxy it to the underlying HttpRequest object.
        """
        try:
            return super(SocketRequest, self).__getattribute__(attr)
        except AttributeError:
            info = sys.exc_info()
            try:
                return getattr(self._message, attr)
            except AttributeError:
                six.reraise(info[0], info[1], info[2].tb_next)

    @property
    def TEXT(self):
        if not _hasattr(self, '_text'):
            self._load_text_and_bytes()
        return self._text

    @property
    def BYTES(self):
        if not _hasattr(self, '_bytes'):
            self._load_text_and_bytes()
        return self._bytes



# coding=utf-8
from __future__ import unicode_literals

import six
from django.db import models
from functools import update_wrapper
from rest_framework import exceptions
from rest_framework import status
from rest_framework.compat import set_rollback

from rest_channels import exceptions as rest_exceptions
from rest_channels.settings import rest_channels_settings
from rest_channels.socket_request import ContentType, SocketRequest


def exception_handler(exc, context):
    if isinstance(exc, exceptions.APIException):
        if isinstance(exc.detail, dict):
            data = exc.detail
        else:
            data = {'detail': exc.detail}
        set_rollback()
        return data
    # Note: Unhandled exceptions will raise a 500 error.
    return None


class BaseChannelsView(object):
    # The following policies may be set at either globally, or per-view.
    text_renderer_class = rest_channels_settings.DEFAULT_TEXT_RENDERER_CLASS
    bytes_renderer_class = rest_channels_settings.DEFAULT_BYTES_RENDERER_CLASS

    text_parser_class = rest_channels_settings.DEFAULT_TEXT_PARSER_CLASS
    bytes_parser_class = rest_channels_settings.DEFAULT_BYTES_PARSER_CLASS

    settings = rest_channels_settings

    def __init__(self, **kwargs):
        """
        Constructor. Called in the URLconf; can contain helpful extra
        keyword arguments, and other things.
        """
        # Go through keyword arguments, and either save their values to our
        # instance, or raise an error.
        for key, value in six.iteritems(kwargs):
            setattr(self, key, value)

    @classmethod
    def as_view(cls, **initkwargs):
        if isinstance(getattr(cls, 'queryset', None), models.query.QuerySet):
            def force_evaluation():
                raise RuntimeError(
                    'Do not evaluate the `.queryset` attribute directly, '
                    'as the result will be cached and reused between requests. '
                    'Use `.all()` or call `.get_queryset()` instead.'
                )

            cls.queryset._fetch_all = force_evaluation
            cls.queryset._result_iter = force_evaluation  # Django <= 1.5

        for key in initkwargs:
            if key in cls.socket_channel_names:
                raise TypeError("You tried to pass in the %s method name as a "
                                "keyword argument to %s(). Don't do that."
                                % (key, cls.__name__))
            if not hasattr(cls, key):
                raise TypeError("%s() received an invalid keyword %r. as_view "
                                "only accepts arguments that are already "
                                "attributes of the class." % (cls.__name__, key))

        def view(message, *args, **kwargs):
            self = cls(**initkwargs)
            self.message = message
            self.args = args
            self.kwargs = kwargs
            return self.dispatch(message, *args, **kwargs)

        view.view_class = cls
        view.view_initkwargs = initkwargs

        update_wrapper(view, cls, updated=())
        update_wrapper(view, cls.dispatch, assigned=())
        view.cls = cls
        return view

    def _allowed_channels(self):
        return [channel_name for channel_name in self.socket_channel_names
                if hasattr(self, channel_name)]

    @property
    def allowed_channels(self):
        return self._allowed_channels()

    def not_allowed(self, request, *args, **kwargs):
        """
        If `message.method` does not correspond to a handler method,
        determine what kind of exception to raise.
        """
        raise rest_exceptions.EventNotAllowed(request.channel.name)

    def get_exception_handler_context(self):
        """
        Returns a dict that is passed through to EXCEPTION_HANDLER,
        as the `context` argument.
        """
        return {
            'view': self,
            'args': getattr(self, 'args', ()),
            'kwargs': getattr(self, 'kwargs', {}),
            'request': getattr(self, 'request', None)
        }

    def get_view_description(self, html=False):
        """
        Return some descriptive text for the view, as used in OPTIONS responses
        and in the browsable API.
        """
        func = self.settings.VIEW_DESCRIPTION_FUNCTION
        return func(self.__class__, html)

    def initialize_request(self, message, *args, **kwargs):
        """
        Returns the initial request object.
        """
        return SocketRequest(
            message,
            text_parser_class=self.text_parser_class,
            bytes_parser_class=self.bytes_parser_class
        )

    def get_renderer(self):
        if self.request.content_type == ContentType.text:
            return self.text_renderer_class
        else:
            return self.bytes_renderer_class

    def get_rendered_data(self, data):
        renderer = self.get_renderer()
        rendered_data = renderer().render(data=data)
        return rendered_data

    def handle_exception(self, exc):
        exception_handler = self.settings.EXCEPTION_HANDLER

        context = self.get_exception_handler_context()
        response_data = exception_handler(exc, context)

        if response_data is None:
            response_data = {
                'detail': 'Internal server error.',
                'status': status.HTTP_500_INTERNAL_SERVER_ERROR
            }
        else:
            status_code = getattr(exc, 'status_code', None)
            if status_code is not None:
                response_data.update({'status': status_code})
        self.send_exception(response_data)

    def send_exception(self, response_data):
        raise NotImplementedError

    def dispatch(self, message, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs
        request = self.initialize_request(message, *args, **kwargs)
        self.request = request
        try:
            channel_name = request.channel.name
            if channel_name in self.channel_names:
                handler = getattr(self, self.channels_map[channel_name], self.not_allowed)
            else:
                handler = self.not_allowed
            handler(request, *args, **kwargs)
        except Exception as exc:
            self.handle_exception(exc)


class WebSocketView(BaseChannelsView):
    channel_names = ('websocket.connect', 'websocket.receive', 'websocket.disconnect')
    channels_map = {
        'websocket.connect': 'connect',
        'websocket.receive': 'receive',
        'websocket.disconnect': 'disconnect'
    }

    def send(self, channel_or_group, data, content_type='text', close=False):
        channel_or_group.send({content_type: self.get_rendered_data(data), 'close': close})

    def send_exception(self, response_data):
        self.send(self.request.reply_channel, response_data)


class EmailView(BaseChannelsView):
    channel_names = ('email.receive',)
    channels_map = {
        'email.receive': 'receive',
    }


class UDPView(BaseChannelsView):
    channel_names = ('udp.receive',)
    channels_map = {
        'udp.receive': 'receive',
    }


class HTTPView(BaseChannelsView):
    channel_names = ('http.request',)
    channels_map = {
        'http.request': 'receive'
    }



# coding=utf-8
from __future__ import unicode_literals

from django.utils.encoding import force_text
from django.utils.translation import ugettext_lazy as _
from rest_framework import status


class ChannelsException(Exception):
    status_code = status.HTTP_500_INTERNAL_SERVER_ERROR
    default_detail = _('A server error occurred.')

    def __init__(self, detail=None):
        if detail is not None:
            self.detail = force_text(detail)
        else:
            self.detail = force_text(self.default_detail)

    def __str__(self):
        return self.detail


class EventNotAllowed(ChannelsException):
    status_code = status.HTTP_405_METHOD_NOT_ALLOWED
    default_detail = _('Event "{event}" not allowed.')

    def __init__(self, event, detail=None):
        if detail is not None:
            self.detail = force_text(detail)
        else:
            self.detail = force_text(self.default_detail).format(event=event)



# coding=utf-8
from __future__ import unicode_literals

from rest_framework import serializers
from rest_framework.exceptions import ValidationError


# noinspection PyProtectedMember,PyAbstractClass
class RouteSerializer(serializers.Serializer):
    method = serializers.CharField()
    data = serializers.DictField()

    def validate_method(self, method_name):
        view = self.context['view']
        method = getattr(view, method_name, None)
        if method is None or not getattr(method, 'socket_route', None):
            raise ValidationError(
                'Method {method_name} not supplied'.format(method_name=method_name)
            )
        return method_name


# noinspection PyAbstractClass
class RouteResponseSerializer(RouteSerializer):
    status = serializers.IntegerField(min_value=100, max_value=600)



import os

import re
from setuptools import find_packages, setup


def get_version(package):
    """
    Return package version as listed in `__version__` in `init.py`.
    """
    init_py = open(os.path.join(package, '__init__.py')).read()
    return re.search("__version__ = ['\"]([^'\"]+)['\"]", init_py).group(1)


# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='rest_channels',
    version=get_version('rest_channels'),
    packages=find_packages(),
    include_package_data=True,
    license='BSD License',
    description='Class Based Views for Django Channels with Django Rest Framework',
    setup_requires=['setuptools-markdown'],
    install_requires=[
        'django>=1.8',
        'djangorestframework>=3.0.0',
        'channels>=0.11.0',
        'ujson>=1.35',
        'six>=1.10.0',
    ],
    long_description_markdown_filename='README.md',
    author='Khasanov Bulat',
    author_email='afti@yandex.ru',
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Topic :: Internet :: WWW/HTTP',
    ],
    url='https://github.com/KhasanovBI/rest_channels'
)



# coding=utf-8
from __future__ import unicode_literals

import ujson


class BaseRenderer(object):
    def render(self, data):
        raise NotImplementedError('Renderer class requires .render() to be implemented')


class UJSONRenderer(BaseRenderer):
    def render(self, data, *args, **kwargs):
        if data is None:
            return ''
        ret = ujson.dumps(data)
        return ret



# coding=utf-8
from __future__ import unicode_literals

from channels.sessions import channel_session
from django.utils.decorators import method_decorator
from rest_channels.socket_routing.serializers import RouteSerializer, RouteResponseSerializer
from rest_channels.views import WebSocketView
from rest_framework import status as rest_framework_status


class SocketRouteView(WebSocketView):
    @method_decorator(channel_session)
    def receive(self, request, *args, **kwargs):
        serializer = RouteSerializer(data=request.data, context={'view': self})
        serializer.is_valid(raise_exception=True)
        method = serializer.data['method']
        self.action = method
        getattr(self, method)(request, serializer.data.get('data'), *args, **kwargs)

    def route_send(self, channel_or_group, data, status=rest_framework_status.HTTP_200_OK,
                   method=None, content_type='text', close=False):
        if method is None:
            assert self.action is not None
            method = self.action
        self.send(
            channel_or_group,
            RouteResponseSerializer({'data': data, 'status': status, 'method': method}).data,
            content_type,
            close
        )

    def handle_exception(self, exc):
        exception_handler = self.settings.EXCEPTION_HANDLER

        context = self.get_exception_handler_context()
        response_data = exception_handler(exc, context)

        if response_data is None:
            self.route_send(
                self.request.reply_channel,
                data={'detail': 'Internal server error.'},
                status=rest_framework_status.HTTP_500_INTERNAL_SERVER_ERROR
            )
        else:
            status_code = getattr(exc, 'status_code', None)
            if status_code is None:
                status_code=rest_framework_status.HTTP_500_INTERNAL_SERVER_ERROR,
            self.route_send(
                self.request.reply_channel,
                data=response_data,
                status=status_code
            )

